
from enum import Enum, unique
import inspect
import re
import os

from lily_assistant.config import Config

from lily.base import serializers, parsers
from lily.base.models import JSONSchemaValidator


class MissingReturnStatementError(Exception):
    """Raise when return value is not declared.

    When Method used for calculating dynamical serialization field is
    missing the return annotation.

    """


class MissingMethodForFieldError(Exception):
    """Raise when method does not exist.

    When Method used for calculating dynamical serialization field is
    missing.

    """


class MissingSchemaMappingError(Exception):
    """Raise when schema mapping is missing.

    When mapping between any type of a field in the Serializer and schema
    is missing.

    """


class ForbiddenFieldError(Exception):
    """Raise when schema field is forbidden."""


@unique
class SerializerType(Enum):

    RESPONSE = 'RESPONSE'

    REQUEST_QUERY = 'REQUEST_QUERY'

    REQUEST_BODY = 'REQUEST_BODY'

    REQUEST_HEADER = 'REQUEST_HEADER'


class SchemaRenderer:

    def __init__(self, serializer, parser_type=None):

        # -- figure out with which type of Serializer or Parser we're
        # -- dealing with
        if inspect.isclass(serializer):
            instance = serializer()

        else:
            instance = serializer

        if isinstance(instance, serializers.Serializer):
            self.type = SerializerType.RESPONSE.value

        elif (isinstance(instance, parsers.Parser) and
                parser_type == parsers.ParserTypeEnum.BODY.value):
            self.type = SerializerType.REQUEST_BODY.value

        elif (isinstance(instance, parsers.Parser) and
                parser_type == parsers.ParserTypeEnum.QUERY.value):
            self.type = SerializerType.REQUEST_QUERY.value

        else:
            self.type = None

        self.serializer = serializer

        # -- fill in the meta data
        self.meta = self.get_meta(serializer)

    def get_meta(self, serializer):
        path = inspect.getfile(serializer)
        path = path.replace(Config.get_project_path(), '')

        return {
            'first_line': inspect.getsourcelines(serializer)[1],
            'path': path,
        }

    def render(self):

        return self.serializer_to_schema(self.serializer)

    def serializer_to_schema(self, serializer):

        schema = Schema(
            self.type,
            self.meta,
            getattr(serializer, '_type', None))

        try:
            fields = serializer.get_fields()

        except TypeError:
            fields = serializer().get_fields()

        for name, field in fields.items():

            # -- list serializer
            if isinstance(field, serializers.ListSerializer):
                schema.add_array(
                    name=name,
                    required=field.child.required,
                    value=self.serializer_to_schema(
                        field.child.__class__()))

            # -- singleton nested serializer
            elif (
                    isinstance(field, serializers.Serializer) or
                    isinstance(field, parsers.Parser)):

                schema.add(
                    name=name,
                    required=field.required,
                    value=self.serializer_to_schema(field))

            # -- method serializer
            elif isinstance(field, serializers.SerializerMethodField):
                schema = self.method_field_to_schema(
                    serializer, name, field, schema)

            # -- normal field boolean field
            elif self.is_simple_field(field):
                value, enums = self.simple_field_to_schema(name, field)
                schema.add(
                    name=name,
                    required=field.required,
                    value=value)
                schema.enums.extend(enums)

            # -- list field
            elif isinstance(field, serializers.ListField):

                if self.is_simple_field(field.child):
                    value, enums = self.simple_field_to_schema(
                        name, field.child)
                    schema.add_array(
                        name=name,
                        required=field.required,
                        value=value)
                    schema.enums.extend(enums)

                else:
                    schema.add_array(
                        name=name,
                        required=field.required,
                        value=self.serializer_to_schema(field.child))

            # -- all not covered cases should raise an Error
            else:
                raise MissingSchemaMappingError(
                    '{serializer}.{name}'.format(
                        serializer=self.get_name(serializer),
                        name=name))

        return schema

    def method_field_to_schema(self, serializer, name, field, schema):

        try:
            method = getattr(serializer, 'get_{}'.format(name))
            return_value = method.__annotations__['return']

        except KeyError:
            raise MissingReturnStatementError(
                'Missing Return Statement for Serializer {0} '
                'and method {1}'.format(self.get_name(serializer), method))

        except AttributeError:
            raise MissingMethodForFieldError(
                'Missing Method Statement for Serializer {0} '
                'and field {1}'.format(self.get_name(serializer), name))

        else:
            if isinstance(return_value, list):
                item_type = return_value[0]

                # -- if it's already a schema
                if isinstance(item_type, dict):
                    schema.add_array(
                        name=name,
                        required=True,
                        value=item_type)

                elif isinstance(item_type(), serializers.Serializer):
                    schema.add_array(
                        name=name,
                        required=True,
                        value=self.serializer_to_schema(item_type))

                else:
                    schema.add_array(
                        name=name,
                        required=True,
                        value=self.native_type_to_schema(item_type))

            # -- if it's already a schema
            elif isinstance(return_value, dict):
                schema.add(
                    name=name,
                    required=True,
                    value=return_value)

            elif isinstance(return_value(), serializers.Serializer):
                schema.add(
                    name=name,
                    required=True,
                    value=self.serializer_to_schema(return_value))

            else:
                schema.add(
                    name=name,
                    required=True,
                    value=self.native_type_to_schema(return_value))

        return schema

    def get_name(self, obj):
        try:
            return obj.__name__

        except AttributeError:
            return obj.__class__.__name__

    def is_simple_field(self, field):
        return isinstance(
            field,
            (
                serializers.BooleanField,
                serializers.CharField,
                serializers.ChoiceField,
                serializers.DateField,
                serializers.DateTimeField,
                serializers.DecimalField,
                serializers.DictField,
                serializers.EmailField,
                serializers.FloatField,
                serializers.IntegerField,
                serializers.JSONField,
                serializers.JSONSchemaField,
                serializers.URLField,
                serializers.ReadOnlyField,
                serializers.UUIDField,
            ))

    def simple_field_to_schema(self, name, field):
        enums = []
        out = None

        def is_field(_types):
            return isinstance(field, _types)

        if is_field(serializers.BooleanField):
            out = {
                'type': 'boolean',
            }

        # FIXME: figure out how to fetch the type hidden behind ReadOnly
        # Field!!!!!!!!!!!!!
        elif is_field(serializers.ReadOnlyField):
            out = {
                'type': 'any',
            }

        # -- must be before CharField
        elif is_field(serializers.EmailField):
            out = {
                'type': 'string',
                'format': 'email',
            }

        # -- must be before CharField
        elif is_field(serializers.URLField):
            out = {
                'type': 'string',
                'format': 'uri',
            }

        elif is_field(serializers.DateField):
            out = {
                'type': 'string',
                'format': 'date',
            }

        elif is_field(serializers.DateTimeField):
            out = {
                'type': 'string',
                'format': 'date-time',
            }

        elif is_field(serializers.UUIDField):
            out = {
                'type': 'string',
            }

        elif is_field(serializers.CharField):
            field_schema = {
                'type': 'string',
            }

            if field.min_length:
                field_schema['minLength'] = field.min_length

            if field.max_length:
                field_schema['maxLength'] = field.max_length

            out = field_schema

        elif is_field(serializers.IntegerField):
            field_schema = {
                'type': 'integer',
            }

            if field.min_value:
                field_schema['minimum'] = field.min_value

            if field.max_value:
                field_schema['maximum'] = field.max_value

            out = field_schema

        elif is_field(serializers.DecimalField):
            out = {
                'type': 'string',
            }

        elif is_field(serializers.FloatField):
            field_schema = {
                'type': 'number',
            }

            if field.min_value:
                field_schema['minimum'] = field.min_value

            if field.max_value:
                field_schema['maximum'] = field.max_value

            out = field_schema

        # -- choice fields are FORBIDDEN since they lead to ambiguous Enums
        # -- on the client side
        elif (
                is_field(serializers.ChoiceField) and
                not is_field(serializers.EnumChoiceField)):
            raise ForbiddenFieldError(
                f'{name} uses ChoiceField which is forbidden since it leads '
                f'to ambiguous client side Enums. Please use '
                f'`EnumChoiceField` instead')

        elif is_field(serializers.EnumChoiceField):
            choices = list(field.choices.keys())
            choice_type = type(choices[0])
            has_same_type = all([isinstance(c, choice_type) for c in choices])

            if has_same_type and isinstance(choices[0], int):
                schema_type = 'integer'

            else:
                schema_type = 'string'

            out = {
                'type': schema_type,
                'enum_name': field.enum_name,
                'enum': sorted(choices),
            }
            enums.append(out)

        elif is_field(serializers.DictField):
            out = {
                'type': 'object',
            }

        elif is_field((serializers.JSONField, serializers.JSONSchemaField)):

            out = {
                'type': 'any',
            }

            # -- search for the 1st JSONSchemaValidator validator
            # -- the should be always at most one.
            for validator in field.validators:
                if isinstance(validator, JSONSchemaValidator):
                    out = validator.schema
                    if 'schemas' in out:
                        out = {
                            'oneOf': out['oneOf']
                        }

        return out, enums

    def native_type_to_schema(self, _type):

        if _type in [int, float]:
            return {
                'type': 'number',
            }

        elif _type == str:
            return {
                'type': 'string',
            }

        elif _type == bool:
            return {
                'type': 'boolean',
            }


class ArrayValue:

    def __init__(self, value):
        self.value = value


class Schema:

    def __init__(self, type_, meta, entity_type=None):
        self.type = type_
        self.meta = meta
        self.schema = self.get_empty_schema()

        if entity_type:
            self.schema['entity_type'] = entity_type

        self.enums = []

    def is_empty(self):
        return self.schema == self.get_empty_schema()

    def get_empty_schema(self):
        return {
            'type': 'object',
            'required': [],
            'properties': {},
        }

    #
    # Adding Values to the Root Schema
    #
    def add_array(self, name, value, required=True):
        if isinstance(value, Schema):
            self.enums.extend(value.enums)

        self.add(name, ArrayValue(value), required)

    def add(self, name, value, required=True):
        if isinstance(value, Schema):
            self.enums.extend(value.enums)

        self.schema['properties'][name] = value
        if required:
            self.schema['required'].append(name)

    #
    # Representation
    #
    def serialize(self):

        def serialize(d):

            entity_type = d.get('entity_type', None)

            o = {
                'type': d['type'],
                'required': d['required'],
                'properties': {},
            }
            if entity_type:
                o['entity_type'] = entity_type

            p = o['properties']

            for k, v in d['properties'].items():
                if isinstance(v, Schema):
                    p[k] = serialize(v.schema)

                elif isinstance(v, ArrayValue):
                    if isinstance(v.value, Schema):
                        p[k] = {
                            'type': 'array',
                            'items': serialize(v.value.schema),
                        }

                    else:
                        p[k] = {
                            'type': 'array',
                            'items': v.value,
                        }

                else:
                    p[k] = v

            return o

        if self.is_empty():
            schema = {}

        else:
            schema = serialize(self.schema)

        return {
            'uri': self.get_repository_uri(),
            'schema': schema,
        }

    def get_repository_uri(self):
        config = Config()

        if 'bitbucket' in config.repository:
            return os.path.join(
                config.repository,
                'src',
                config.last_commit_hash,
                re.sub(r'^/', '', self.meta['path']),
                '#lines-{}'.format(self.meta['first_line']))

        elif 'github':
            return os.path.join(
                config.repository,
                'blob',
                config.last_commit_hash,
                re.sub(r'^/', '', self.meta['path']),
                '#L{}'.format(self.meta['first_line']))




from unittest.mock import call
import json

from django.core import management
from django.test import TestCase
from click.testing import CliRunner
import pytest

from migrator.management.commands.apply_migration_plan_for_version import (
    command,
)


class ApplyMigrationPlanForVersionCommandTestCase(TestCase):

    @pytest.fixture(autouse=True)
    def init_fixtures(self, mocker, tmpdir):
        self.mocker = mocker
        self.tmpdir = tmpdir

    def setUp(self):
        self.runner = CliRunner()

        class MockConfig:

            @classmethod
            def get_lily_path(cls):
                return str(self.tmpdir)

        self.mocker.patch(
            'migrator.management.commands.apply_migration_plan_for_version'
            '.Config',
            MockConfig)

    def test_apply_migrations(self):

        call_command = self.mocker.patch.object(management, 'call_command')

        self.tmpdir.mkdir('migrations').join('6.7.1.json').write(json.dumps({
            'plan': [
                ('auth', '0001_a'),
                ('media', '0002_b'),
            ],
        }))

        result = self.runner.invoke(command, ['6.7.1'])

        assert result.exit_code == 0
        assert result.output.strip() == (
            'Migrations plan for version 6.7.1 applied.')
        assert call_command.call_args_list == [
            call('migrate', 'auth', '0001_a'),
            call('migrate', 'media', '0002_b'),
        ]




from __future__ import unicode_literals

from django.db import migrations
from django.db.migrations.operations.base import Operation


class CreateFunctionAndAggregate(Operation):

    reversible = True

    def __init__(self, func_name, func_body, agg_name, agg_body):
        self.func_name = func_name
        self.func_body = func_body
        self.agg_name = agg_name
        self.agg_body = agg_body

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor, *args, **kwargs):
        schema_editor.execute(
            "CREATE FUNCTION %s %s" % (self.func_name, self.func_body))
        schema_editor.execute(
            "CREATE AGGREGATE %s %s" % (self.agg_name, self.agg_body))

    def database_backwards(self, app_label, schema_editor, *args, **kwargs):
        schema_editor.execute("DROP AGGREGATE %s" % self.agg_name)
        schema_editor.execute("DROP FUNCTION %s" % self.func_name)

    def describe(self):
        return (
            "Creates function {0} and aggregate {1}".format(
                self.func_name, self.agg_name))


class Migration(migrations.Migration):

    dependencies = [
        ('search', '0001_add_polish_language_support'),
    ]

    operations = [
        CreateFunctionAndAggregate(
            # function to be used in aggregate data
            "tsvector_add(tsvector, tsvector)",
            """
                RETURNS tsvector
                AS 'SELECT COALESCE($1, '''') || COALESCE($2, '''');'
                LANGUAGE SQL;
            """,
            # aggregate data
            "tsvector_agg(tsvector)",
            """
            (
                sfunc = tsvector_add,
                stype = tsvector
            )
            """),
    ]




import os
import textwrap

import click

from .copier import Copier
from .logger import Logger


logger = Logger()


"""
Quick hack to force lily_assistant to see correct encoding locales.

Please keep in mind that those encoding will live only in the process
spawned by lily_assistant and will not contaminate your global namespace.

"""
if not os.environ.get('LC_ALL'):  # pragma: no cover
    os.environ['LC_ALL'] = 'en_US.utf-8'


if not os.environ.get('LANG'):  # pragma: no cover
    os.environ['LANG'] = 'en_US.utf-8'


@click.group()
def cli():
    """Expose multiple commands allowing one to work with lily_assistant."""

    pass


@click.command()
@click.argument('src_dir')
def init(src_dir):
    """Init `lily` with some preliminary files copying and creation.

    Init `Lily`. During this operation the following will take place:
    - `.lily/lily.makefile` will be copied to the root project directory.

    WARNING it is assumed that this command will be invoked in the root
    of the project.

    :param src_dir: name of your source directory

    """

    # -- check that init for lily_assistant was executed too
    if not os.path.exists(os.path.join(os.getcwd(), '.lily', 'config.json')):
        raise click.ClickException(textwrap.dedent('''
            Please install `lily_assistant` and run
            `lily_assistant init <src_dir>` before running
            `lily init <src_dir>`
        '''))

    Copier().copy(src_dir)

    logger.info('''

        Please insert the following line at the top of your Makefile:

        include .lily/lily.makefile
    ''')


cli.add_command(init)




from django.test import TestCase

from search.stopwords import stopwords_filter


class StopWordsFilterTestCase(TestCase):

    def test_is_stopword__polish(self):

        assert stopwords_filter.is_stopword(
            'polish', 'jak') is True
        assert stopwords_filter.is_stopword(
            'polish', 'ale') is True
        assert stopwords_filter.is_stopword(
            'polish', 'albo') is True
        assert stopwords_filter.is_stopword(
            'polish', 'wszędzie') is False
        assert stopwords_filter.is_stopword(
            'polish', 'cześć') is False

    def test_is_stopword__non_polish(self):

        assert stopwords_filter.is_stopword(
            'english', 'hi') is False
        assert stopwords_filter.is_stopword(
            'english', 'and') is False
        assert stopwords_filter.is_stopword(
            'english', 'or') is False
        assert stopwords_filter.is_stopword(
            'english', 'what') is False




from django.conf import settings


SECRET_KEY = 'not.really.needed'


DEBUG = True


ALLOWED_HOSTS = ["*"]


#
# LILY SPECIFIC SETTINGS
#
LILY_AUTHORIZER_CLASS = getattr(
    settings,
    'LILY_AUTHORIZER_CLASS',
    'lily.base.authorizer.BaseAuthorizer')


LILY_AUTHORIZER_ACCESS_ENUM_CLASS = getattr(
    settings,
    'LILY_AUTHORIZER_ACCESS_ENUM_CLASS',
    'AccountType')


LILY_ENTRYPOINT_COMMANDS_ACCESS_LIST = getattr(
    settings,
    'LILY_ENTRYPOINT_COMMANDS_ACCESS_LIST',
    None)


LILY_EXCLUDE_QUERY_PARSER_ALL_OPTIONAL_ASSERTIONS = getattr(
    settings,
    'LILY_EXCLUDE_QUERY_PARSER_ALL_OPTIONAL_ASSERTIONS',
    None)

# FIXME: !!!! should be moved to github!
LILY_ANGULAR_CLIENT_ORIGIN = (
    'git@bitbucket.org:goodai/lily-angular-client.git')
# git@bitbucket.org:goodai/lily-angular-client-base.git

# FIXME: !!!! should be moved to github!
LILY_ANGULAR_HTTP_ORIGIN = (
    'git@bitbucket.org:goodai/lily-angular-http.git')

#
# ASYNC
#
LILY_ASYNC_LOCK_DB_HOST = getattr(
    settings,
    'LILY_ASYNC_LOCK_DB_HOST',
    'localhost')

LILY_ASYNC_LOCK_DB_PORT = getattr(
    settings,
    'LILY_ASYNC_LOCK_DB_PORT',
    6379)

LILY_ASYNC_LOCK_DB_INDEX = getattr(
    settings,
    'LILY_ASYNC_LOCK_DB_INDEX',
    1)

#
# Internationalization
#
LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


#
# Apps viewed and maintained by both Admin and API
#
INSTALLED_APPS = [
    'rest_framework',

    'docs',
    'base',
    'search',
    'entrypoint',
    'assertion',
    'migrator',

    'django.contrib.contenttypes',
    'django.contrib.auth',
]


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'lily',
        'USER': 'local',
        'PASSWORD': '',
        'HOST': '127.0.0.1',
        'PORT': '5434',
    },
}


MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
]


ROOT_URLCONF = getattr(
    settings,
    'ROOT_URLCONF',
    'conf.urls')


WSGI_APPLICATION = getattr(
    settings,
    'WSGI_APPLICATION',
    'conf.wsgi.application')


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [],
        },
    },
]




from django.contrib.postgres.search import SearchRank


class Rank(SearchRank):
    """Rank for the calculation of frequency match with 2 normalizations.

    - type 16 normalization - divides the rank by 1 + the logarithm of the
      number of unique words in document
    - type 32 normalization - divides the rank by itself + 1

    """

    def as_sql(self, compiler, connection, function=None, template=None):
        return super(Rank, self).as_sql(
            compiler,
            connection,
            function='ts_rank',
            template='%(function)s(%(expressions)s, 16 | 32)')


class RankCD(SearchRank):
    """Rank for the calculation of harmonic mean distances between tokens.

    It uses Cover Density Rank function and type 4 normalization - divides
    the rank by the mean harmonic distance between extents

    """

    def as_sql(self, compiler, connection, function=None, template=None):
        return super(RankCD, self).as_sql(
            compiler,
            connection,
            function='ts_rank_cd',
            template='%(function)s(%(expressions)s, 4)')




from django.test import TestCase
import pytest

from lily.base.meta import Meta, MetaSerializer, Domain, DomainSerializer
from lily.base.events import EventFactory


class MetaTestCase(TestCase):

    def test_required_fields(self):
        # -- missing title
        with pytest.raises(TypeError):
            Meta(domain=Domain(id='d', name='d'))

        # -- missing domain
        with pytest.raises(TypeError):
            Meta(title='hi there')

    def test_arguments_are_saved(self):

        m = Meta(
            title='hi there',
            description='this is it',
            domain=Domain(id='d', name='d'))

        assert m.title == 'hi there'
        assert m.description == 'this is it'
        assert m.domain == Domain(id='d', name='d')


class MetaSerializerTestCase(TestCase):

    def test_serialization(self):

        m = Meta(
            title='hi there',
            description='this is it',
            domain=Domain(id='a', name='aaa'))

        assert MetaSerializer(m).data == {
            '@type': 'meta',
            'title': 'hi there',
            'description': 'this is it',
            'domain': {
                '@type': 'domain',
                'id': 'a',
                'name': 'aaa',
            },
        }


class DomainTestCase(TestCase):

    def test_required_fields(self):
        # -- missing id
        with pytest.raises(TypeError):
            Domain(name='hi')

        # -- missing name
        with pytest.raises(TypeError):
            Domain(id='hi')

    def test_arguments_are_saved(self):

        d = Domain(id='cards', name='Cards Management')

        assert d.id == 'cards'
        assert d.name == 'Cards Management'

    def test_invalid_id(self):

        with pytest.raises(EventFactory.BrokenRequest) as e:
            Domain(id='cards manager', name='...')

        assert e.value.event == 'BROKEN_ARGS_DETECTED'
        assert e.value.data == {
            '@event': 'BROKEN_ARGS_DETECTED',
            '@type': 'error',
            'errors': {
                'id': [
                    'should not contain white characters.',
                ]
            }
        }


class DomainSerializerTestCase(TestCase):

    def test_serialization(self):

        d = Domain(
            id='super',
            name='Super Management')

        assert DomainSerializer(d).data == {
            '@type': 'domain',
            'id': 'super',
            'name': 'Super Management',
        }


@pytest.mark.parametrize(
    'description, expected',
    [
        # -- empty to empty
        (
            '',
            '',
        ),

        # -- nothing to transform
        (
            'hello world',
            'hello world',
        ),

        # -- prefix white chars
        (
            '\t\t   hello\n world',
            'hello world',
        ),

        # -- suffix white chars
        (
            'hello\n world \t\t\n  \n',
            'hello world',
        ),

        # -- single space delimiter
        (
            'hello\nworld',
            'hello world',
        ),
    ])
def test_transform_description(description, expected):

    m = Meta(title='...', domain=Domain(id='d', name='d'))

    assert m.transform_description(description) == expected




from enum import Enum, unique

from rest_framework import serializers as drf_serializers

# -- needed for correct namespacing so that one could access all
# -- serializers stuff as a part of the `parsers` namespace
from .serializers import *  # noqa


@unique
class ParserTypeEnum(Enum):

    BODY = 'BODY'

    QUERY = 'QUERY'

    HEADER = 'HEADER'


class Parser(drf_serializers.Serializer):
    """Parses all types of payload: body, query, headers (future)."""

    @classmethod
    def as_query_parser(cls, *args, **kwargs):
        return cls(*args, parser_type=ParserTypeEnum.QUERY.value, **kwargs)

    @classmethod
    def as_body_parser(cls, *args, **kwargs):
        return cls(*args, parser_type=ParserTypeEnum.BODY.value, **kwargs)

    def __init__(
            self,
            *args,
            parser_type=ParserTypeEnum.BODY.value,
            **kwargs):

        raw_data = kwargs.get('data')
        if raw_data is None:
            kwargs['data'] = {}

        if parser_type == ParserTypeEnum.QUERY.value:
            kwargs['data'] = self.prepare_query_data(raw_data)

        elif parser_type == ParserTypeEnum.BODY.value:
            kwargs['data'] = self.prepare_body_data(raw_data)

        super(Parser, self).__init__(*args, **kwargs)

    def prepare_query_data(self, raw_data):

        data = {}
        if raw_data is not None:
            for field_name, field_class in self.fields.items():

                if not isinstance(field_class, ListField):  # noqa
                    try:
                        data[field_name] = raw_data[field_name]

                    # -- ignore key errors since they all will be caught
                    # -- by the validation
                    except (KeyError, IndexError):
                        pass

                else:
                    try:
                        data[field_name] = raw_data.getlist(field_name)

                    except AttributeError:
                        data[field_name] = raw_data.get(field_name)

        return data

    def prepare_body_data(self, raw_data):

        return raw_data


class PageParser(Parser):

    offset = IntegerField(default=0)  # noqa

    limit = IntegerField(default=100)  # noqa


class FullTextSearchParser(Parser):

    query = CharField(default=None)  # noqa


class ModelParser(drf_serializers.ModelSerializer, Parser):

    serializer_choice_field = EnumChoiceField  # noqa

    def build_standard_field(self, field_name, model_field):

        from . import models  # noqa - avoid circular dependency

        field_class, field_kwargs = super(
            ModelParser, self).build_standard_field(
                field_name, model_field)

        if isinstance(model_field, models.EnumChoiceField):
            field_kwargs['enum_name'] = model_field.enum_name
            field_class = EnumChoiceField  # noqa

        return field_class, field_kwargs
